"""remarx: Marx quote identification for the CDH Citing Marx project."""

# Copyright (c) 2026, Center for Digital Humanities, Princeton University
# SPDX-License-Identifier: Apache-2.0

from remarx import app, quotation, sentence

__version__ = "1.0.1"

__all__ = ["__version__", "app", "quotation", "sentence"]
