"""
This module contains libraries for embedding generation and quotation detection.
"""

from remarx.quotation import consolidate, embeddings, pairs

__all__ = ["consolidate", "embeddings", "pairs"]
